/*
 * Copyright 1991 University of Wisconsin-Madison
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Wisconsin-Madison not
 * be used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The University of
 * Wisconsin-Madison makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *
 * THE UNIVERSITY OF WISCONSIN-MADISON DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF WISCONSIN-MADISON BE LIABLE FOR
 * ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:   Marvin Solomon
 *           Computer Sciences Department
 *           University of Wisconsin, Madison WI
 *           solomon@cs.wisc.edu or seismo!uwvax!solomon
 *
 * Current Maintainer:
 *
 *           Tim Theisen           Systems Programmer
 * Internet: tim@cs.wisc.edu       Department of Computer Sciences
 *     UUCP: uwvax!tim             University of Wisconsin-Madison
 *    Phone: (608)262-0438         1210 West Dayton Street
 *      FAX: (608)262-9777         Madison, WI   53706
 */
/*
 * $Log: patchlevel.h,v $
 * Revision 1.8  1991/10/04  22:05:44  tim
 * Add copyright and permission notice.
 *
 * Revision 1.7  1991/10/03  20:38:15  tim
 * Modify Imakefile for R5.
 *   - Font compiler now generates .pcf files.
 * Add DEPLIBS to Imakefile so xproof builds better with xmkmf.
 *
 * Add minor type casts in Proof.c to match function prototypes.
 *
 * Added code to main to restrict the size of the window if it
 * will be bigger than the screen.
 *
 * Revision 1.6  1990/08/04  10:23:55  tim
 * Patches from Andreas Stolcke <stolcke@icsi.Berkeley.EDU>
 *
 * -  allow Bezier curves (D'z ...') to be faked using splines, instead of
 *    complaining about an undefined command. Bezier curves are not available
 *   in all ditroff versions, but it doesn't hurt putting them in.  They are
 *    generated by the gremlin version in use here at Berkeley.  If you feel
 *    like implementing real Bezier curves in xproof let me know.  I can send
 *    you the relevant code in gr2ps or psdit that does it.
 *
 * -  xproof originally doesn't care about the 'x T ...' command at the beginning
 *    of the input.  I modified it so it set's its 'device' variable accordingly.
 *    Also, the DEVICE macro wasn't used to define the default device.
 *
 * -  The #define InstalledLibraries business in Imakefile is no longer necessary
 *    in R4 since the R4 imake rules already provide such an option
 *    (generating the Makefile with xmkmf causes the installed libs to be used).
 *
 * Revision 1.5  90/06/25  19:15:42  tim
 * Redid caching of font_id's.  They don't belong in the font table.
 * If recorded in the font table, only one size of each type face could
 * be loaded.
 * 
 * Revision 1.4  90/06/22  16:29:49  tim
 * Changed NoPixmap to UsePixmap.  Put UsePixmap under user control.
 * The usePixmap resource defaults to true for all Architectures except
 * VaxArchitecture (for Xqdss) and BobcatArchitecture.  The resource
 * can be overridden in the resource database or on the command line.
 * The command line option -usepixmap enables usePixmap.  The command
 * line option +usepixmap disables usePixmap.  (I am following the
 * convention used by xterm, et. al.)
 * 
 * Support 75dpi and 100dpi fonts.
 * 
 * Use screen resolution to determine which fonts to use.
 * 
 * Added code to check if file has changed whenever displaying a new page.
 * If changed the file is reopened.
 * 
 * Added a refresh button, which cause redisplay of the current page.
 * 
 * Recorded font_id's with in font_map table.  Only call XLoadQueryFont
 * once for each font.  (Wild carded match within the server can take
 * significant time to determine whether font is loaded.)
 * 
 * Made font string more specific to speed up search for font name.
 * 
 * Revision 1.3  90/03/05  13:41:11  tim
 * Athena widget changes.  Mostly includes have moved to X11/Xaw.
 * VPaned nolonger exists. Using Paned instead.
 * 
 * Revision 1.2  89/07/17  11:01:01  solomon
 * fixfont.c:      Added #endif.
 * getdevinfo.c:
 *         Fixed off-by-one error:  Size-numbers range from 0 to nsizes
 *         *inclusive*.  The size[] table was right, but the xfont[] structures
 *         were wrong.  Also fixed header comment to correctly document
 * 		dev.nfonts, nfaces, and maxmount.
 * showpage.c:  Ditto(LookupSize() was missing the final entry).
 * xfontinfo.c:	UnloadXFonts() fixed to run through the face[] array (all faces
 * 		"seen") rather than the mounted[] array (faces currently "mounted").
 * 
 * Revision 1.1  89/07/17  08:52:39  solomon
 * Initial revision
 * 
 */
#define PATCHLEVEL 0
